from __future__ import absolute_import, division, print_function

import numpy as np
from qtpy import QtGui
from glue.core import roi as roimod

__all__ = ['cmap2pixmap', 'ginga_graphic_to_roi']


def cmap2pixmap(cmap, steps=50):
    """Convert a Ginga colormap into a QtGui.QPixmap

    :param cmap: The colormap to use
    :type cmap: Ginga colormap instance (e.g. ginga.cmap.get_cmap('gray'))
    :param steps: The number of color steps in the output. Default=50
    :type steps: int

    :rtype: QtGui.QPixmap
    """
    inds = np.linspace(0, 1, steps)
    n = len(cmap.clst) - 1
    tups = [cmap.clst[int(x * n)] for x in inds]
    rgbas = [QtGui.QColor(int(r * 255), int(g * 255),
                          int(b * 255), 255).rgba() for r, g, b in tups]
    im = QtGui.QImage(steps, 1, QtGui.QImage.Format_Indexed8)
    im.setColorTable(rgbas)
    for i in range(steps):
        im.setPixel(i, 0, i)
    im = im.scaled(128, 32)
    pm = QtGui.QPixmap.fromImage(im)
    return pm


def ginga_graphic_to_roi(obj):
    if obj.kind == 'rectangle':
        # make sure x2, y2 are the upper right corner
        x1, y1, x2, y2 = obj.swapxy(obj.x1, obj.y1, obj.x2, obj.y2)
        roi = roimod.RectangularROI(xmin=x1, xmax=x2, ymin=y1, ymax=y2)
    elif obj.kind == 'circle':
        roi = roimod.CircularROI(xc=obj.x, yc=obj.y,
                                     radius=obj.radius)
    elif obj.kind in ('polygon', 'freepolygon'):
        vx, vy = zip(*obj.points)
        roi = roimod.PolygonalROI(vx=vx, vy=vy)

    elif obj.kind in ('path', 'freepath', 'line'):
        vx, vy = zip(*obj.points)
        roi = roimod.Path(vx=vx, vy=vy)

    elif obj.kind == 'xrange':
        x1, y1, x2, y2 = obj.swapxy(obj.x1, 0, obj.x2, 0)
        roi = roimod.XRangeROI(min=x1, max=x2)

    elif obj.kind == 'yrange':
        x1, y1, x2, y2 = obj.swapxy(0, obj.y1, 0, obj.y2)
        roi = roimod.YRangeROI(min=y1, max=y2)

    elif obj.kind == 'point':
        roi = roimod.PointROI(x=obj.x, y=obj.y)

    else:
        raise Exception("Don't know how to convert shape '%s' to a ROI" % (
            obj.kind))

    return roi



#
# Glue.py -- Glue plugin for Ginga viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
The Glue plugin implements a Glue interface for the Ginga viewer.
"""
from __future__ import absolute_import, division, print_function

import sys
import warnings

from astropy.table import Table
from astropy.wcs import WCS

from ginga import GingaPlugin
from ginga import AstroImage
from ginga.gw import Widgets
from ginga.misc.Datasrc import Datasrc
from ginga.misc.Callback import Callbacks
from ginga.table import AstroTable
from ginga.util.six.moves import map

from glue.core import Data
from glue.core.coordinates import WCSCoordinates
from glue.core.message import (DataCollectionAddMessage,
                               DataCollectionDeleteMessage,
                               ApplicationClosedMessage)
from glue.core.hub import HubListener
from glue.utils.error import GlueDeprecationWarning

help_msg = sys.modules[__name__].__doc__

__all__ = ['Glue']


class GingaHubListener(Callbacks, HubListener):
    """
    Acts as a bridge between the Glue data collection hub and ginga callbacks.
    """

    def __init__(self, datasrc):
        super(GingaHubListener, self).__init__()

        self.datasrc = datasrc

        for cbname in ['data_in', 'data_out', 'app_closed']:
            self.enable_callback(cbname)

    def connect_hub(self, hub):
        hub.subscribe(self, DataCollectionAddMessage,
                      self._data_added_cb)
        hub.subscribe(self, DataCollectionDeleteMessage,
                      self._data_removed_cb)

        # This needs https://github.com/glue-viz/glue/pull/1168
        hub.subscribe(self, ApplicationClosedMessage,
                      self._app_closed_cb)

    def _data_added_cb(self, msg):
        data = msg.data
        name = data.label
        if data.ndim == 1:
            image = self._data_to_table(data)
        else:
            image = self._data_to_image(data)
        image.set(name=name)
        self.datasrc[name] = image
        self.make_callback('data_in', image)

    def _data_removed_cb(self, msg):
        data = msg.data
        name = data.label
        image = self.datasrc.remove(name)
        self.make_callback('data_out', image)

    def _app_closed_cb(self, msg):
        self.make_callback('app_closed', None)

    def get_data(self, name):
        return self.datasrc[name]

    @staticmethod
    def _data_to_image(data):
        ids = data.component_ids()
        data_np = data[ids[0]]
        data_meta = {}

        if hasattr(data.coords, 'header'):
            h = AstroImage.AstroHeader()
            h.update(data.coords.header)
            data_meta['header'] = h

        image = AstroImage.AstroImage(data_np=data_np, metadata=data_meta)

        if hasattr(data.coords, 'wcs'):
            image.wcs.load_header(data.coords.wcs.to_header())

        return image

    @staticmethod
    def _data_to_table(data):
        tab = Table()

        for cid in data.visible_components:
            comp = data.get_component(cid)
            if comp.categorical:
                tab[cid.label] = comp.labels
            else:
                tab[cid.label] = comp.data

        return AstroTable.AstroTable(data_ap=tab)


class Glue(GingaPlugin.GlobalPlugin):
    """Glue global plugin for Ginga reference viewer."""

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(Glue, self).__init__(fv)

        self.glue_app = None
        self.glue_hl = GingaHubListener(Datasrc(length=0))
        self.glue_hl.add_callback('data_in', self.data_added_cb)
        self.glue_hl.add_callback('data_out', self.data_removed_cb)
        self.glue_hl.add_callback('app_closed', self.app_closed_cb)

        self.data_names = []
        self.gui_up = False

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)

        vbox, sw, orientation = Widgets.get_oriented_box(container)

        vbox.set_border_width(4)
        vbox.set_spacing(2)

        self.msg_font = self.fv.get_font("sansFont", 12)
        tw = Widgets.TextArea(wrap=True, editable=False)
        tw.set_font(self.msg_font)
        self.tw = tw

        fr = Widgets.Expander("Instructions")
        fr.set_widget(tw)
        vbox.add_widget(fr, stretch=0)

        fr = Widgets.Frame("Glue Interface")

        captions = [
            ('Start Glue', 'button', 'Stop Glue', 'button'),
            ('Put Data', 'button'),
            ('Get Data', 'button', 'dataitem', 'combobox'),
            ]
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        fr.set_widget(w)
        vbox.add_widget(fr, stretch=0)

        b.start_glue.add_callback('activated', lambda w: self.start_glue_cb())
        b.start_glue.set_tooltip('Start a Glue session')

        b.stop_glue.add_callback('activated', lambda w: self.stop_glue_cb())
        b.stop_glue.set_tooltip('Stop a Glue session')
        b.stop_glue.set_enabled(False)

        b.put_data.add_callback('activated', lambda w: self.put_data_cb())
        b.put_data.set_tooltip('Send data to Glue')
        b.put_data.set_enabled(False)

        b.get_data.add_callback('activated', lambda w: self.get_data_cb())
        b.get_data.set_tooltip('Get selected data from Glue')
        b.get_data.set_enabled(False)

        b.dataitem.set_tooltip('Select data to get from Glue')

        # stretch
        top.add_widget(sw, stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(4)
        btns.set_border_width(4)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btn.set_tooltip('Close this plugin and Glue')
        btns.add_widget(btn)
        btns.add_widget(Widgets.Label(''), stretch=1)

        top.add_widget(btns, stretch=0)
        container.add_widget(top, stretch=1)

        self.gui_up = True

    def instructions(self):
        self.tw.set_text("""This plugin enables interface with Glue.

Press "Start Glue" to start a new Glue session. Glue started independently (without using this button) does not work with this plugin. Glue started this way is tied to the Ginga session; i.e., closing Ginga will also close Glue.

Press "Stop Glue" to end the Glue session without closing the plugin. This might not work correctly if there are multiple Glue sessions. This is unnecessary if Glue is already closed, e.g., by pressing "X" in Glue.

To send an image or table to Glue, make it the currently active image/table and press "Put Data". Then switch to the Glue application to interact with it.

To get an image or table from Glue to the currently active channel, select the associated name from the drop-down menu and then press "Get Data". If there is already an image with the same name in the Ginga channel, it will be overwritten.

Press "Close" to close this plugin. This also closes the associated Glue session, if not already.""")  # noqa

    def error_no_glue(self, verbose=True):
        """Call this to reset GUI when Glue session disappears."""
        self.w.start_glue.set_enabled(True)
        self.w.stop_glue.set_enabled(False)
        self.w.put_data.set_enabled(False)
        self.w.get_data.set_enabled(False)
        self.w.dataitem.clear()
        self.data_names = []
        if verbose:
            self.fv.show_error("No glue session running!")

    def put_data_cb(self):
        if self.glue_app is None:
            self.error_no_glue()
            return

        channel = self.fv.get_current_channel()
        if channel is None:
            self.fv.show_error("No active channel!")
            return

        image = channel.get_current_image()
        if image is None:
            self.fv.show_error("No data in channel '%s'" % (channel.name))
            return

        data_np = image.get_data()
        name = image.get('name', 'noname')
        kwargs = {name: data_np}

        try:
            # Pass in WCS for image.
            if isinstance(image, AstroImage.AstroImage):
                gdata = Data(**kwargs)
                h = image.get_header()
                w = WCS(h)
                gdata.coords = WCSCoordinates(h, wcs=w)
                self.glue_app.add_data(**{name: gdata})
            # Table data
            else:
                self.glue_app.add_data(**kwargs)

            self.glue_app.raise_()

        except Exception as e:
            self.fv.show_error("Error sending data to Glue: %s" % (str(e)))
            self.error_no_glue()

    def get_data_cb(self):
        if self.glue_app is None:
            self.error_no_glue()
            return

        # channel = self.fv.get_channel_on_demand('Glue')
        channel = self.fv.get_current_channel()
        if channel is None:
            self.fv.show_error("No active channel!")
            return

        idx = self.w.dataitem.get_index()
        name = self.data_names[idx]
        dataobj = self.glue_hl.get_data(name)

        channel.add_image(dataobj)

    def _adj_data_list(self):
        # adjust available items list
        dc = self.glue_app.data_collection
        self.w.dataitem.clear()

        self.data_names = list(map(lambda data: data.label, dc.data))

        if len(self.data_names) == 0:
            self.w.get_data.set_enabled(False)
            return

        for name in self.data_names:
            self.w.dataitem.append_text(name)

        self.w.dataitem.set_index(0)
        self.w.get_data.set_enabled(True)

    def data_added_cb(self, hub, dataobj):
        self._adj_data_list()

    def data_removed_cb(self, hub, dataobj):
        self._adj_data_list()

    def app_closed_cb(self, hub, dataobj):
        self.glue_app = None
        self.error_no_glue(verbose=False)

    def start_glue_cb(self):
        self.glue_app = qglue()
        hub = self.glue_app.data_collection.hub
        if hub is not None:
            self.glue_hl.connect_hub(hub)

        # self.glue_app._create_terminal()
        sgeo = self.glue_app.app.desktop().screenGeometry()
        self.glue_app.show()
        self.glue_app.has_terminal()
        self.glue_app.resize(sgeo.width() * 0.9, sgeo.height() * 0.9)
        # self.glue_app.lower()

        # Toggle buttons accordingly.
        self.w.start_glue.set_enabled(False)
        self.w.stop_glue.set_enabled(True)
        self.w.put_data.set_enabled(True)

    def stop_glue_cb(self):
        if self.glue_app is None:
            self.error_no_glue(verbose=False)
            return

        w, self.glue_app = self.glue_app, None

        try:
            w.deleteLater()
        except Exception as e:  # Glue is closed already
            self.logger.debug(str(e))
        finally:
            self.error_no_glue(verbose=False)

    def glue_new_data_cb(self, msg):
        print(dir(msg))

    def start(self):
        self.instructions()

        if self.glue_app is not None:
            self.w.start_glue.set_enabled(False)
            self.w.stop_glue.set_enabled(True)
            self.w.put_data.set_enabled(True)
            self._adj_data_list()

    def stop(self):
        self.gui_up = False
        # self.stop_glue_cb()  # Uncomment this to close Glue when plugin close

    def close(self):
        self.w.dataitem.clear()

        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'glue'


def qglue():
    """
    Quickly send python variables to Glue for visualization.

    Returns
    -------
    ga : ``GlueApplication`` object

    """
    from glue.core import DataCollection
    from glue.app.qt import GlueApplication

    dc = DataCollection()

    # Suppress pesky Glue warnings.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', GlueDeprecationWarning)
        ga = GlueApplication(data_collection=dc)

    return ga

# END



from __future__ import absolute_import, division, print_function

from glue.utils.qt import get_qapp  # noqa

# The application has to always be referenced to avoid being shut down, so we
# keep a reference to it here
app = None


def pytest_configure(config):
    global app
    app = get_qapp()



from .version import __version__  # noqa


def setup():

    try:
        import ginga  # noqa
    except ImportError:
        raise ImportError("ginga is required")

    from .qt import mouse_modes  # noqa
    from .qt.viewer_widget import GingaViewer
    from glue.config import qt_client
    qt_client.add(GingaViewer)



from __future__ import absolute_import, division, print_function

import os

from qtpy import QtWidgets, PYQT5

from ginga.misc import log
from ginga import toolkit

if PYQT5:
    toolkit.use('qt5')
else:
    toolkit.use('qt')

from ginga.gw import ColorBar
from ginga.gw import Readout

from ginga.qtw.ImageViewCanvasQt import ImageViewCanvas
from ginga.misc.Settings import SettingGroup
from ginga.Bindings import ImageViewBindings
from ginga.util.paths import ginga_home
from ginga import colors

from glue.viewers.common.qt.data_viewer_with_state import DataViewerWithState
from glue.viewers.common.qt.toolbar import BasicToolbar
from glue.viewers.image.state import ImageViewerState
from glue.viewers.image.qt.options_widget import ImageOptionsWidget
from glue.core import command
from glue.core.edit_subset_mode import EditSubsetMode

# The following is to ensure that the mouse modes get registered
from glue_ginga.qt import mouse_modes  # noqa

from glue_ginga.qt.layer_artist import GingaImageLayer, GingaSubsetImageLayer

__all__ = ['GingaViewer']


class GingaViewer(DataViewerWithState):

    LABEL = "Ginga Viewer"

    _toolbar_cls = BasicToolbar
    _state_cls = ImageViewerState
    _options_cls = ImageOptionsWidget
    _data_artist_cls = GingaImageLayer
    _subset_artist_cls = GingaSubsetImageLayer

    tools = ['ginga:rectangle', 'ginga:circle', 'ginga:polygon', 'ginga:lasso',
             'ginga:xrange', 'ginga:yrange', 'ginga:pan', 'ginga:freepan',
             'ginga:rotate', 'ginga:contrast', 'ginga:cuts', 'ginga:dist',
             'ginga:colormap', 'ginga:spectrum', 'ginga:slicer']

    def __init__(self, session, parent=None, state=None):

        super(GingaViewer, self).__init__(session, parent, state=state)

        self.logger = log.get_logger(name='ginga', level=20,
                                     # switch commenting for debugging
                                     null=True, log_stderr=False,
                                     #null=False, log_stderr=True
                                     )

        # load binding preferences if available
        cfgfile = os.path.join(ginga_home, "bindings.cfg")
        bindprefs = SettingGroup(name='bindings', logger=self.logger,
                                 preffile=cfgfile)
        bindprefs.load(onError='silent')

        bd = ImageViewBindings(self.logger, settings=bindprefs)

        # make Ginga viewer
        self.viewer = ImageViewCanvas(self.logger, render='widget',
                                      bindings=bd)
        self.canvas = self.viewer

        # prevent widget from grabbing focus
        self.viewer.set_enter_focus(False)
        self.viewer.set_desired_size(300, 300)

        # enable interactive features
        bindings = self.viewer.get_bindings()
        bindings.enable_all(True)
        self.canvas.register_for_cursor_drawing(self.viewer)
        self.canvas.add_callback('draw-event', self._apply_roi_cb)
        self.canvas.add_callback('edit-event', self._update_roi_cb)
        self.canvas.add_callback('draw-down', self._clear_roi_cb)
        self.canvas.enable_draw(False)
        self.canvas.enable_edit(False)
        self.viewer.enable_autozoom('off')
        self.viewer.set_zoom_algorithm('rate')
        self.viewer.set_zoomrate(1.4)
        self.viewer.set_fg(*colors.lookup_color("#D0F0E0"))

        bm = self.viewer.get_bindmap()
        bm.add_callback('mode-set', self.mode_set_cb)
        self.mode_w = None
        self.mode_actns = {}

        # Create settings and set defaults
        settings = self.viewer.get_settings()
        self.settings = settings
        settings.getSetting('cuts').add_callback('set', self.cut_levels_cb)
        settings.set(autozoom='off', autocuts='override',
                     autocenter='override')

        # make color bar, with color maps shared from ginga canvas
        rgbmap = self.viewer.get_rgbmap()
        self.colorbar = ColorBar.ColorBar(self.logger)
        rgbmap.add_callback('changed', self.rgbmap_cb, self.viewer)
        self.colorbar.set_rgbmap(rgbmap)

        # make coordinates/value readout
        self.readout = Readout.Readout(-1, 20)
        self.roi_tag = None
        self.opn_obj = None

        topw = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.viewer.get_widget(), stretch=1)
        cbar_w = self.colorbar.get_widget()
        if not isinstance(cbar_w, QtWidgets.QWidget):
            # ginga wrapped widget
            cbar_w = cbar_w.get_widget()
        layout.addWidget(cbar_w, stretch=0)
        readout_w = self.readout.get_widget()
        if not isinstance(readout_w, QtWidgets.QWidget):
            # ginga wrapped widget
            readout_w = readout_w.get_widget()
        layout.addWidget(readout_w, stretch=0)
        topw.setLayout(layout)

        self._crosshair_id = '_crosshair'

        self.setCentralWidget(topw)

    def get_layer_artist(self, cls, layer=None, layer_state=None):
        if layer_state is not None and layer_state.viewer_state is None:
            layer_state.viewer_state = self.state
        return cls(layer=layer, layer_state=layer_state,
                   canvas=self.canvas, viewer_state=self.state)

    def match_colorbar(self, canvas, colorbar):
        rgbmap = self.viewer.get_rgbmap()
        loval, hival = self.viewer.get_cut_levels()
        colorbar.set_range(loval, hival)
        colorbar.set_rgbmap(rgbmap)

    def rgbmap_cb(self, rgbmap, canvas):
        self.match_colorbar(canvas, self.colorbar)

    def cut_levels_cb(self, setting, tup):
        (loval, hival) = tup
        self.colorbar.set_range(loval, hival)

    def _set_roi_mode(self, opn_obj, name, mode, **kwargs):
        self.opn_obj = opn_obj
        en_draw = (mode == 'draw')
        self.canvas.enable_draw(en_draw)
        self.canvas.set_draw_mode(mode)
        # XXX need better way of setting draw contexts
        self.canvas.draw_context = self
        self.canvas.set_drawtype(name, **kwargs)

    def _clear_roi_cb(self, canvas, *args):
        if self.opn_obj is not None:
            self.opn_obj.opn_init(self, self.roi_tag)
        else:
            try:
                self.canvas.delete_object_by_tag(self.roi_tag)
            except:
                pass

    def _apply_roi_cb(self, canvas, tag):
        if self.canvas.draw_context is not self:
            return
        self.roi_tag = tag
        obj = self.canvas.get_object_by_tag(tag)

        if self.opn_obj is None:
            # delete outline
            self.canvas.delete_object(obj)
            self.roi_tag = None
            return

        self.opn_obj.opn_exec(self, tag, obj)

    def _update_roi_cb(self, canvas, obj):
        if self.canvas.draw_context is not self:
            return
        if self.opn_obj is None:
            return
        self.opn_obj.opn_update(self, obj)

    def mode_cb(self, modname, tf):
        """This method is called when a toggle button in the toolbar is pressed
        selecting one of the modes.
        """
        bm = self.viewer.get_bindmap()
        if not tf:
            bm.reset_mode(self.viewer)
            return
        bm.set_mode(modname, mode_type='locked')
        return True

    def mode_set_cb(self, bm, modname, mtype):
        """This method is called when a mode is selected in the viewer widget.
        NOTE: it may be called when mode_cb() is not called (for example, when
        a keypress initiates a mode); however, the converse is not true: calling
        mode_cb() will always result in this method also being called as a result.

        This logic is to insure that the toggle buttons are left in a sane state
        that reflects the current mode, however it was initiated.
        """
        if modname in self.mode_actns:
            if self.mode_w and (self.mode_w != self.mode_actns[modname]):
                self.mode_w.setChecked(False)
            self.mode_w = self.mode_actns[modname]
            self.mode_w.setChecked(True)
        elif self.mode_w:
            # keystroke turned on a mode for which we have no GUI button
            # and a GUI button is selected--unselect it
            self.mode_w.setChecked(False)
            self.mode_w = None
        return True

    def set_cmap(self, cmap):
        self.canvas.set_cmap(cmap)

    def show_crosshairs(self, x, y):
        self.clear_crosshairs()
        c = self.canvas.viewer.get_draw_class('point')(x, y, 6, color='red', style='plus')
        self.canvas.add(c, tag=self._crosshair_id, redraw=True)

    def clear_crosshairs(self):
        try:
            self.canvas.delete_objects_by_tag([self._crosshair_id], redraw=False)
        except:
            pass

    def apply_roi(self, roi):
        cmd = command.ApplyROI(data_collection=self._data,
                               roi=roi, apply_func=self._apply_roi)
        self._session.command_stack.do(cmd)

    def _apply_roi(self, roi):

        x_comp = self.state.x_att.parent.get_component(self.state.x_att)
        y_comp = self.state.y_att.parent.get_component(self.state.y_att)

        subset_state = x_comp.subset_from_roi(self.state.x_att, roi,
                                              other_comp=y_comp,
                                              other_att=self.state.y_att,
                                              coord='x')

        mode = EditSubsetMode()
        mode.update(self._data, subset_state)



from __future__ import absolute_import, division, print_function

import os
import sys

from ginga import cmap as ginga_cmap

from qtpy import QtGui, QtWidgets
from glue.config import viewer_tool
from glue.viewers.common.qt.tool import CheckableTool
from glue.utils import nonpartial
from glue.utils.qt import load_ui
from glue.plugins.tools.spectrum_tool.qt import SpectrumTool
from glue.plugins.tools.pv_slicer.qt import PVSlicerMode

from glue_ginga.qt.utils import cmap2pixmap, ginga_graphic_to_roi

# Find out location of ginga module so we can some of its icons
GINGA_HOME = os.path.split(sys.modules['ginga'].__file__)[0]
GINGA_ICON_DIR = os.path.join(GINGA_HOME, 'icons')

# add matplotlib colormaps
# TODO: glue menu seems extremely slow if we add them all
#ginga_cmap.add_matplotlib_cmaps(fail_on_import_error=False)


class GingaROIMode(CheckableTool):

    def opn_init(self, viewer, tag):
        """This method is called when the user clicks down to draw an object.
        It gets called with any previously drawn object that was kept (or None).
        """
        if tag is not None:
            try:
                # typical case for a ROI is delete any existing shape
                viewer.canvas.delete_object_by_tag(tag)
            except:
                pass

        return None

    def opn_exec(self, viewer, tag, obj):
        """This method gets called when the user finishes drawing an object
        and it has been added to the canvas.  `tag` is the tag of the object
        on the canvas and `obj` is the actual shape.
        """

        # typical case is we want to remove the shape we just drew
        # from the canvas because we will be replacing it with a ROI
        viewer.canvas.delete_object_by_tag(tag, redraw=False)

        roi = ginga_graphic_to_roi(obj)

        viewer.apply_roi(roi)


class GingaPathMode(GingaROIMode):
    pass


@viewer_tool
class RectangleROIMode(GingaROIMode):

    tool_id = 'ginga:rectangle'
    icon = 'glue_square'
    action_text = 'Rectangular ROI'
    tool_tip = 'Define a rectangular region of interest'

    def activate(self):
        self.viewer._set_roi_mode(self, 'rectangle', 'draw',
                                  color='cyan', linewidth=2, linestyle='dash',
                                  fill=True, fillcolor='yellow', fillalpha=0.5)

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'rectangle', None)


@viewer_tool
class CircleROIMode(GingaROIMode):

    tool_id = 'ginga:circle'
    icon = 'glue_circle'
    action_text = 'Circular ROI'
    tool_tip = 'Define a circular region of interest'

    def activate(self):
        self.viewer._set_roi_mode(self, 'circle', 'draw',
                                  color='cyan', linewidth=2, linestyle='dash',
                                  fill=True, fillcolor='yellow', fillalpha=0.5)

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'circle', None)


@viewer_tool
class PolygonROIMode(GingaROIMode):

    tool_id = 'ginga:polygon'
    icon = 'glue_lasso'
    action_text = 'Polygonal ROI'
    tool_tip = ('Define a polygon region of interest.\n'
                '  Click and drag to start the polygon;\n'
                '    "v" adds a vertex,\n'
                '    "z" deletes last added vertex.\n'
                '  Release button to finalize')
    status_tip = ('CLICK and DRAG to start polygon, press "v" to add a '
                  'polygon vertex, "z" to remove last added vertex')

    def activate(self):
        self.viewer._set_roi_mode(self, 'polygon', 'draw',
                                  color='cyan', linewidth=2, linestyle='dash',
                                  fill=True, fillcolor='yellow', fillalpha=0.5)

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'polygon', None)


@viewer_tool
class LassoROIMode(GingaROIMode):

    tool_id = 'ginga:lasso'
    icon = 'glue_lasso'
    tool_tip = ('Lasso a region of interest\n'
                '  Click and drag in a lasso shape to select;\n'
                '  Release to finalize.')
    action_text = 'Polygonal ROI'
    status_tip = ('CLICK and DRAG to start lassoing')

    def activate(self):
        self.viewer._set_roi_mode(self, 'freepolygon', 'draw',
                                  color='cyan', linewidth=2, linestyle='dash',
                                  fill=True, fillcolor='yellow', fillalpha=0.5)

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'freepolygon', None)


@viewer_tool
class PathROIMode(GingaROIMode):

    tool_id = 'ginga:path'
    icon = 'glue_lasso'
    tool_tip = 'Define a path region of interest'
    tool_tip = ('Draw a path as a region of interest\n'
                '  "v" adds a vertex\n'
                '  "z" deletes last added vertex\n'
                '  release button to finalize')
    status_tip = ('CLICK and DRAG to start path, press "v" to add a '
                  'path vertex, "z" to remove last added vertex')

    def activate(self):
        self.viewer._set_roi_mode(self, 'path', 'draw',
                                  color='cyan', linewidth=2, linestyle='dash')

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'path', None)


@viewer_tool
class HRangeMode(GingaROIMode):
    """
    Defines a Range ROI, accessible via the :meth:`~HRangeMode.roi` method.

    This class defines horizontal ranges
    """

    icon = 'glue_xrange_select'
    tool_id = 'ginga:xrange'
    action_text = 'X range'
    tool_tip = 'Select a range of x values'

    def activate(self):
        self.viewer._set_roi_mode(self, 'xrange', 'draw',
                                  fillcolor='cyan', fillalpha=0.5,
                                  linewidth=0)

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'xrange', None)


@viewer_tool
class VRangeMode(GingaROIMode):
    """
    Defines a Range ROI, accessible via the :meth:`~VRangeMode.roi` method.

    This class defines vertical ranges.
    """

    icon = 'glue_yrange_select'
    tool_id = 'ginga:yrange'
    action_text = 'Y range'
    tool_tip = 'Select a range of y values'

    def activate(self):
        self.viewer._set_roi_mode(self, 'yrange', 'draw',
                                  fillcolor='cyan', fillalpha=0.5,
                                  linewidth=0)

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'yrange', None)


@viewer_tool
class PickMode(GingaROIMode):
    """
    Defines a PointROI.

    Defines single point selections.
    """

    icon = 'glue_yrange_select'
    tool_id = 'ginga:pick'
    action_text = 'Pick'
    tool_tip = 'Select a single item'

    def activate(self):
        self.viewer._set_roi_mode(self, 'point', 'draw')

    def deactivate(self):
        self.viewer._set_roi_mode(None, 'point', None)


@viewer_tool
class PanMode(CheckableTool):

    tool_id = 'ginga:pan'
    icon = 'glue_move'
    tool_tip = ('Pan the image by dragging left button\n'
               '   zoom by dragging right button')
    status_tip = 'CLICK and DRAG (left btn) to pan, (right btn) to zoom '

    def activate(self):
        self.viewer.mode_cb('pan', True)

    def deactivate(self):
        self.viewer.mode_cb('pan', False)


@viewer_tool
class FreePanMode(CheckableTool):

    tool_id = 'ginga:freepan'
    icon = os.path.join(GINGA_ICON_DIR, 'hand_48.png')
    tool_tip = ('Zoom in and set pan by clicking left btn\n'
                '  zoom out and set pan by clicking right btn\n'
                '  click and drag middle btn to pan freely')
    status_tip = 'CLICK to zoom and set pan (in == left btn, out == right btn)'

    def activate(self):
        self.viewer.mode_cb('freepan', True)

    def deactivate(self):
        self.viewer.mode_cb('freepan', False)


@viewer_tool
class RotateMode(CheckableTool):

    tool_id = 'ginga:rotate'
    icon = os.path.join(GINGA_ICON_DIR, 'rotate_48.png')
    tool_tip = 'Click and drag to rotate the image about the pan position'
    status_tip = 'Click and drag to rotate the image about the pan position'

    def activate(self):
        self.viewer.mode_cb('rotate', True)

    def deactivate(self):
        self.viewer.mode_cb('rotate', False)


@viewer_tool
class ContrastMode(CheckableTool):

    tool_id = 'ginga:contrast'
    icon = 'glue_contrast'
    tool_tip = ('Adjust bias/contrast of the image, ds9-style\n'
                '  CLICK and DRAG: horizontally to set bias (shift colormap),\n'
                '    vertically to set contrast (stretch colormap).\n'
                '  CLICK right btn to restore to normal (colormap)')

    def activate(self):
        self.viewer.mode_cb('contrast', True)

    def deactivate(self):
        self.viewer.mode_cb('contrast', False)

    # TODO: uncomment when we have updated Ginga to version that contains
    # the restore_contrast() method

    ## def menu_actions(self):
    ##     result = []

    ##     a = QtWidgets.QAction("Restore", None)
    ##     a.triggered.connect(nonpartial(self.restore_cb))
    ##     result.append(a)

    ##     return result

    def restore_cb(self):
        gviewer = self.viewer.viewer
        gviewer.restore_contrast()


@viewer_tool
class CutsMode(CheckableTool):

    tool_id = 'ginga:cuts'
    icon = os.path.join(GINGA_ICON_DIR, 'cuts_48.png')
    tool_tip = ('Set low and high cut levels of the image interactively\n'
                '  CLICK and DRAG: horizontally to set high cut level,\n'
                '    vertically to set low cut level;\n'
                '    (hold SHIFT to set only low level;\n'
                '     CTRL (Command on Mac) to set only high)\n'
                '  CLICK right button to restore to auto cut levels.\n'
                '  SCROLL to increase/decrease contrast by cut levels;\n'
                '    hold CTRL down to use finer granularity')
    status_tip = ('CLICK and DRAG to set cut levels; horizontally to set low'
                  ' cut level, vertically to set high cut level')

    def activate(self):
        self.viewer.mode_cb('cuts', True)

    def deactivate(self):
        self.viewer.mode_cb('cuts', False)

    def get_vmin_vmax(self):
        gviewer = self.viewer.viewer
        return gviewer.get_cut_levels()

    def set_vmin_vmax(self, vmin, vmax):
        gviewer = self.viewer.viewer
        gviewer.cut_levels(vmin, vmax)

    def do_autocuts(self):
        gviewer = self.viewer.viewer
        gviewer.auto_levels()

    def set_autocuts(self, name):
        gviewer = self.viewer.viewer
        gviewer.set_autocut_params(name)

    def choose_vmin_vmax(self):
        # Following example of glue matplotlib viewer
        from glue.viewers.common.qt import mouse_mode
        dialog = load_ui('contrastlimits.ui', None,
                         directory=os.path.dirname(mouse_mode.__file__))
        v = QtGui.QDoubleValidator()
        dialog.vmin.setValidator(v)
        dialog.vmax.setValidator(v)

        vmin, vmax = self.get_vmin_vmax()
        if vmin is not None:
            dialog.vmin.setText(str(vmin))
        if vmax is not None:
            dialog.vmax.setText(str(vmax))

        def _apply():
            try:
                vmin = float(dialog.vmin.text())
                vmax = float(dialog.vmax.text())
                self.set_vmin_vmax(vmin, vmax)
            except ValueError:
                pass

        bb = dialog.buttonBox
        bb.button(bb.Apply).clicked.connect(_apply)
        dialog.accepted.connect(_apply)
        dialog.show()
        dialog.raise_()
        dialog.exec_()

    def menu_actions(self):
        from ginga.AutoCuts import get_autocuts_names
        result = []

        a = QtWidgets.QAction("autocuts", None)
        a.triggered.connect(nonpartial(self.do_autocuts))
        result.append(a)

        rng = QtWidgets.QAction("Set cuts...", None)
        rng.triggered.connect(nonpartial(self.choose_vmin_vmax))
        result.append(rng)

        a = QtWidgets.QAction("", None)
        a.setSeparator(True)
        result.append(a)

        names = list(get_autocuts_names())
        if 'clip' in names:
            # not relevant to glue
            names.remove('clip')

        for name in names:
            a = QtWidgets.QAction(name, None)
            a.triggered.connect(nonpartial(self.set_autocuts, name))
            result.append(a)

        ## rng = QtWidgets.QAction("Set params...", None)
        ## #rng.triggered.connect(nonpartial(self.choose_vmin_vmax))
        ## result.append(rng)

        return result


@viewer_tool
class DistributionMode(CheckableTool):

    tool_id = 'ginga:dist'
    icon = os.path.join(GINGA_ICON_DIR, 'histogram_48.png')
    tool_tip = ('Adjust value distribution of the image')

    def activate(self):
        self.viewer.mode_cb('dist', True)

    def deactivate(self):
        self.viewer.mode_cb('dist', False)

    def set_dist(self, name):
        gviewer = self.viewer.viewer
        gviewer.set_color_algorithm(name)

    def menu_actions(self):
        from ginga.ColorDist import get_dist_names
        result = []

        for algname in get_dist_names():
            a = QtWidgets.QAction(algname, None)
            a.triggered.connect(nonpartial(self.set_dist, algname))
            result.append(a)

        return result


class ColormapAction(QtWidgets.QAction):

    def __init__(self, label, cmap, parent):
        super(ColormapAction, self).__init__(label, parent)
        self.cmap = cmap
        pm = cmap2pixmap(cmap)
        self.setIcon(QtGui.QIcon(pm))


@viewer_tool
class ColormapMode(CheckableTool):

    icon = 'glue_rainbow'
    tool_id = 'ginga:colormap'
    action_text = 'Set color map'
    tool_tip = 'Set the color map used for the image'

    def activate(self):
        self.viewer.mode_cb('cmap', True)

    def deactivate(self):
        self.viewer.mode_cb('cmap', False)

    def menu_actions(self):
        acts = []
        for label in ginga_cmap.get_names():
            cmap = ginga_cmap.get_cmap(label)
            a = ColormapAction(label, cmap, self.viewer)
            a.triggered.connect(nonpartial(self.viewer.set_cmap, cmap))
            acts.append(a)
        return acts


@viewer_tool
class GingaSpectrumMode(GingaROIMode):

    icon = 'glue_spectrum'
    tool_id = 'ginga:spectrum'
    action_text = 'Spectrum'
    tool_tip = 'Extract a spectrum from the selection'

    def __init__(self, viewer, **kwargs):
        super(GingaSpectrumMode, self).__init__(viewer, **kwargs)

        self._shape_obj = None
        self._shape = 'rectangle'

        self.viewer.state.add_callback('reference_data', self._display_data_hook)

        self._tool = SpectrumTool(self.viewer, self)
        #self._move_callback = self._tool._move_profile

    def _display_data_hook(self, data):
        if data is not None:
            self.enabled = data.ndim == 3

    def menu_actions(self):

        result = []

        a = QtWidgets.QAction('Rectangle', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'rectangle'))
        result.append(a)

        a = QtWidgets.QAction('Circle', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'circle'))
        result.append(a)

        a = QtWidgets.QAction('Lasso', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'freepolygon'))
        result.append(a)

        a = QtWidgets.QAction('Polygon', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'polygon'))
        result.append(a)

        return result

    def set_roi_tool(self, mode):
        self._shape = mode
        self.viewer._set_roi_mode(self, mode, 'draw',
                                  color='red', linewidth=2, linestyle='solid',
                                  fill=False, alpha=1.0)

    def activate(self):
        self.set_roi_tool(self._shape)

    def deactivate(self):
        self.clear()
        self.viewer._set_roi_mode(None, self._shape, None)

    def opn_init(self, viewer, tag):
        self.clear()

    def opn_exec(self, viewer, tag, obj):
        self.clear()
        self._shape_obj = obj

        roi = ginga_graphic_to_roi(obj)
        self._tool._update_from_roi(roi)

    def clear(self):
        if self._shape_obj is not None:
            try:
                self.viewer.canvas.delete_object(self._shape_obj)
            except:
                pass
            self._shape_obj = None

    def close(self):
        self.clear()
        self._tool.close()
        return super(GingaSpectrumMode, self).close()


@viewer_tool
class GingaPVSlicerMode(GingaROIMode):

    icon = 'glue_slice'
    tool_id = 'ginga:slicer'
    action_text = 'Slice Extraction'
    tool_tip = 'Extract a slice from an arbitrary path'

    def __init__(self, viewer, **kwargs):
        super(GingaPVSlicerMode, self).__init__(viewer, **kwargs)

        self._path_obj = None
        self._shape = 'freepath'

        self.viewer.state.add_callback('reference_data', self._display_data_hook)
        self._slice_widget = None

    def _display_data_hook(self, data):
        if data is not None:
            self.enabled = data.ndim == 3

    def menu_actions(self):

        result = []

        a = QtWidgets.QAction('Freepath', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'freepath'))
        result.append(a)

        a = QtWidgets.QAction('Path', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'path'))
        result.append(a)

        return result

    def set_roi_tool(self, mode):
        self._shape = mode
        self.viewer._set_roi_mode(self, mode, 'draw',
                                  color='red', linewidth=2, linestyle='solid',
                                  fill=False, alpha=1.0)

    def _clear_path(self):
        if self._path_obj is not None:
            try:
                self.viewer.canvas.delete_object(self._path_obj)
            except:
                pass
            self._path_obj = None

    def activate(self):
        self.set_roi_tool(self._shape)

    def deactivate(self):
        self._clear_path()
        self.viewer._set_roi_mode(None, self._shape, None)

    _build_from_vertices = PVSlicerMode._build_from_vertices

    def opn_init(self, viewer, tag):
        self._clear_path()

    def opn_exec(self, viewer, tag, obj):
        self._clear_path()
        self._path_obj = obj

        vx, vy = zip(*obj.points)
        self._build_from_vertices(vx, vy)



from __future__ import absolute_import, division, print_function

from time import time

import numpy as np
from ginga.misc import Bunch
from ginga.util import wcsmod
from ginga import AstroImage, BaseImage

from glue.external.echo import keep_in_sync, CallbackProperty
from glue.core.exceptions import IncompatibleAttribute
from glue.core.layer_artist import LayerArtistBase
from glue.utils import color2rgb
from glue.viewers.image.state import ImageLayerState, ImageSubsetLayerState

wcsmod.use('astropy')


class GingaLayerArtist(LayerArtistBase):

    zorder = CallbackProperty()
    visible = CallbackProperty()

    def __init__(self, viewer_state=None, layer=None, layer_state=None, canvas=None):

        super(GingaLayerArtist, self).__init__(layer)

        self._canvas = canvas

        self.layer = layer or layer_state.layer
        self.state = layer_state or self._layer_state_cls(viewer_state=viewer_state,
                                                          layer=self.layer)

        self._viewer_state = viewer_state

        # Should not be needed here? (i.e. should be in add_data/add_subset?)
        if self.state not in self._viewer_state.layers:
            self._viewer_state.layers.append(self.state)

        self.zorder = self.state.zorder
        self.visible = self.state.visible

        self._sync_zorder = keep_in_sync(self, 'zorder', self.state, 'zorder')
        self._sync_visible = keep_in_sync(self, 'visible', self.state, 'visible')

        self.state.add_global_callback(self.update)
        self._viewer_state.add_global_callback(self.update)

    def clear(self):
        self._canvas.delete_objects_by_tag([self._tag], redraw=True)

    def redraw(self, whence=0):
        self._canvas.redraw(whence=whence)

    def remove(self):
        self.clear()

    def __gluestate__(self, context):
        return dict(state=context.id(self.state))

    def update(self, **kwargs):
        try:
            canvas_img = self._canvas.get_object_by_tag(self._tag)
        except KeyError:
            pass
        else:
            canvas_img.set_zorder(self.state.zorder)


class GingaImageLayer(GingaLayerArtist):

    _layer_state_cls = ImageLayerState

    def __init__(self, viewer_state=None, layer=None, layer_state=None, canvas=None):

        super(GingaImageLayer, self).__init__(viewer_state=viewer_state, layer=layer,
                                              layer_state=layer_state, canvas=canvas)

        self._tag = '_image'
        self._img = DataImage(self.state)

    def _ensure_added(self):
        """
        Add artist to canvas if needed
        """
        try:
            self._canvas.get_object_by_tag(self._tag)
        except KeyError:
            self._canvas.set_image(self._img)

    def update(self, **kwargs):

        super(GingaImageLayer, self).update(**kwargs)

        if self.state.visible and self._img:
            self._ensure_added()
        elif not self.state.visible:
            self.clear()
            return

        self.redraw()


class GingaSubsetImageLayer(GingaLayerArtist):

    _layer_state_cls = ImageSubsetLayerState

    def __init__(self, viewer_state=None, layer=None, layer_state=None, canvas=None):

        super(GingaSubsetImageLayer, self).__init__(viewer_state=viewer_state, layer=layer,
                                                    layer_state=layer_state, canvas=canvas)

        self._tag = "layer%s_%s" % (layer.label, time())

        self._img = SubsetImage(self.state)

        # SubsetImages can't be added to canvases directly. Need
        # to wrap into a ginga canvas type.
        Image = self._canvas.get_draw_class('image')
        self._cimg = Image(0, 0, self._img, alpha=0.5, flipy=False)

    def _visible_changed(self, *args):
        if self.state.visible and self._cimg:
            self._canvas.add(self._cimg, tag=self._tag, redraw=True)
        elif not self.state.visible:
            self.clear()

    def _check_enabled(self):
        """
        Sync the enabled/disabled status, based on whether
        mask is computable
        """
        try:
            # Just try computing the subset for the first pixel
            view = tuple(0 for _ in self.layer.data.shape)
            self.layer.to_mask(view)
        except IncompatibleAttribute as exc:
            self.disable_invalid_attributes(*exc.args)
            return

        self.enable()

    def _ensure_added(self):
        """
        Add artist to canvas if needed
        """
        try:
            self._canvas.get_object_by_tag(self._tag)
        except KeyError:
            self._canvas.add(self._cimg, tag=self._tag, redraw=False)

    def update(self, **kwargs):

        super(GingaSubsetImageLayer, self).update(**kwargs)

        self._check_enabled()

        if self.state.visible and self._img:
            self._ensure_added()
        elif not self.state.visible:
            self.clear()
            return

        self.redraw(whence=0)


def forbidden(*args):
    raise ValueError("Forbidden")


class DataImage(AstroImage.AstroImage):
    """
    A Ginga image subclass to interface with Glue Data objects
    """

    get_data = _get_data = copy_data = set_data = get_array = transfer = forbidden

    def __init__(self, layer_state, **kwargs):
        """
        Parameters
        ----------
        ...
        kwargs : dict
            Extra kwargs are passed to the superclass
        """
        self.layer_state = layer_state
        super(DataImage, self).__init__(**kwargs)

    @property
    def shape(self):
        """
        The shape of the 2D view into the data
        """
        return self.layer_state.get_sliced_data_shape()

    def _get_fast_data(self):
        return self._slice((slice(None, None, 10), slice(None, None, 10)))

    def _slice(self, view):
        """
        Extract a view from the 2D image.
        """
        return self.layer_state.get_sliced_data(view=view)


class SubsetImage(BaseImage.BaseImage):
    """
    A Ginga image subclass to interface with Glue subset objects
    """
    get_data = _get_data = copy_data = set_data = get_array = transfer = forbidden

    def __init__(self, layer_state=None, **kwargs):
        """
        Parameters
        ----------
        ...
        kwargs : dict
            Extra kwargs are passed to the ginga superclass
        """
        self.layer_state = layer_state

        # NOTE: BaseImage accesses shape property--we need above items
        # defined because we override shape()
        super(SubsetImage, self).__init__(**kwargs)
        self.order = 'RGBA'

    @property
    def shape(self):
        """
        Shape of the 2D view into the subset mask
        """
        return self.layer_state.get_sliced_data_shape()

    def _rgb_from_mask(self, mask):
        """
        Turn a boolean mask into a 4-channel RGBA image
        """
        r, g, b = color2rgb(self.layer_state.color)
        ones = mask * 0 + 255
        alpha = mask * 127
        result = np.dstack((ones * r, ones * g, ones * b, alpha)).astype(np.uint8)
        return result

    def _get_fast_data(self):
        return self._slice((slice(None, None, 10), slice(None, None, 10)))

    def _calc_order(self, order):
        # Override base class because it invokes a glue forbidden method to
        # access the data type of the image--we can instead assume RGBA
        self.order = 'RGBA'

    def _slice(self, view):
        """
        Extract a view from the 2D subset mask.
        """
        try:
            return self._rgb_from_mask(self.layer_state.get_sliced_data(view=view))
        except IncompatibleAttribute:
            return np.zeros(self.shape + (4,))

    def _set_minmax(self):
        # we already know the data bounds
        self.minval = 0
        self.maxval = 256
        self.minval_noinf = self.minval
        self.maxval_noinf = self.maxval

    def get_scaled_cutout_wdht(self, x1, y1, x2, y2, new_wd, new_ht):
        doit = getattr(self, '_doit', False)
        self._doit = not doit

        # default implementation if downsampling
        if doit or new_wd <= (x2 - x1 + 1) or new_ht <= (y2 - y1 + 1):
            return super(SubsetImage, self).get_scaled_cutout_wdht(x1, y1, x2, y2,
                                                                   new_wd, new_ht)

        # if upsampling, prevent extra to_mask() computation
        x1, x2 = np.clip([x1, x2], 0, self.width - 2).astype(np.int)
        y1, y2 = np.clip([y1, y2], 0, self.height - 2).astype(np.int)

        result = self._slice(np.s_[y1:y2 + 1, x1:x2 + 1])

        yi = np.linspace(0, result.shape[0], new_ht).astype(np.int).reshape(-1, 1).clip(0, result.shape[0] - 1)
        xi = np.linspace(0, result.shape[1], new_wd).astype(np.int).reshape(1, -1).clip(0, result.shape[1] - 1)
        yi, xi = [np.array(a) for a in np.broadcast_arrays(yi, xi)]
        result = result[yi, xi]

        scale_x = 1.0 * result.shape[1] / (x2 - x1 + 1)
        scale_y = 1.0 * result.shape[0] / (y2 - y1 + 1)

        return Bunch.Bunch(data=result, scale_x=scale_x, scale_y=scale_y)



from __future__ import absolute_import, division, print_function

import pytest

pytest.importorskip('ginga')

from glue.viewers.common.qt.tests.test_data_viewer import BaseTestDataViewer

from ..viewer_widget import GingaViewer


class TestGingaViewer(BaseTestDataViewer):
    ndim = 2
    widget_cls = GingaViewer
